@@ -40,6 +40,10 @@
  * 5-10-2002
  * Remove __isdigit and use new ctype.h version.
  * Add conditional setting of QUAL_CHARS for size_t and ptrdiff_t.
+ *
+ * 8-16-2002
+ * Fix two problems that showed up with the python 2.2.1 tests; one
+ *    involving %o and one involving %f.
  */
 
 
@@ -61,6 +65,7 @@
 #include <printf.h>
 
 #ifdef __STDIO_THREADSAFE
+#include <stdio_ext.h>
 #include <pthread.h>
 #endif /* __STDIO_THREADSAFE */
 
@@ -108,7 +113,7 @@
 #define __PA_FLAG_INTMASK \
 	(__PA_FLAG_CHAR|PA_FLAG_SHORT|__PA_FLAG_INT|PA_FLAG_LONG|PA_FLAG_LONG_LONG)
 
-extern printf_function *_custom_printf_handler[MAX_USER_SPEC];
+extern printf_function _custom_printf_handler[MAX_USER_SPEC];
 extern printf_arginfo_function *_custom_printf_arginfo[MAX_USER_SPEC];
 extern char *_custom_printf_spec;
 
@@ -177,13 +182,13 @@ enum {
  */
 
 /* TODO -- Fix the table below to take into account stdint.h. */
-#ifndef LLONG_MAX
-#error fix QUAL_CHARS for no long long!  Affects 'L', 'j', 'q', 'll'.
-#else
-#if LLONG_MAX != INTMAX_MAX
-#error fix QUAL_CHARS intmax_t entry 'j'!
-#endif
-#endif
+/*  #ifndef LLONG_MAX */
+/*  #error fix QUAL_CHARS for no long long!  Affects 'L', 'j', 'q', 'll'. */
+/*  #else */
+/*  #if LLONG_MAX != INTMAX_MAX */
+/*  #error fix QUAL_CHARS intmax_t entry 'j'! */
+/*  #endif */
+/*  #endif */
 
 #ifdef PDS
 #error PDS already defined!
@@ -191,6 +196,9 @@ enum {
 #ifdef SS
 #error SS already defined!
 #endif
+#ifdef IMS
+#error IMS already defined!
+#endif
 
 #if PTRDIFF_MAX == INT_MAX
 #define PDS		0
@@ -212,11 +220,21 @@ enum {
 #error fix QUAL_CHARS size_t entries 'z', 'Z'!
 #endif
 
+#if INTMAX_MAX == INT_MAX
+#define IMS		0
+#elif INTMAX_MAX == LONG_MAX
+#define IMS		4
+#elif defined(LLONG_MAX) && (INTMAX_MAX == LLONG_MAX)
+#define IMS		8
+#else
+#error fix QUAL_CHARS ptrdiff_t entry 't'!
+#endif
+
 #define QUAL_CHARS		{ \
 	/* j:(u)intmax_t z:(s)size_t  t:ptrdiff_t  \0:int */ \
 	/* q:long_long  Z:(s)size_t */ \
 	'h',   'l',  'L',  'j',  'z',  't',  'q', 'Z',  0, \
-	 2,     4,    8,    8,   SS,  PDS,    8,  SS,   0, /* TODO -- fix!!! */\
+	 2,     4,    8,  IMS,   SS,  PDS,    8,  SS,   0, /* TODO -- fix!!! */\
      1,     8 \
 }
 
@@ -311,7 +329,7 @@ extern int _ppfs_parsespec(ppfs_t *ppfs); /* parses specifier */
  */
 
 size_t parse_printf_format(register const char *template,
-						   size_t n, int *argtypes)
+						   size_t n, register int *argtypes)
 {
 	ppfs_t ppfs;
 	size_t i;
@@ -751,7 +769,7 @@ extern int _ppfs_parsespec(ppfs_t *ppfs)
 	int n;
 	int argtype[MAX_ARGS_PER_SPEC+2];
 	int argnumber[3];			/* width, precision, 1st data arg */
-	unsigned int conv_num;
+	unsigned int conv_num;		/* This does not need to be initialized. */
 	static const char spec_flags[] = SPEC_FLAGS;
 	static const char spec_chars[] = SPEC_CHARS;/* TODO: b? */
 	static const char spec_ranges[] = SPEC_RANGES;
@@ -1021,7 +1039,7 @@ int register_printf_function(int spec, printf_function handler,
 /**********************************************************************/
 #ifdef L__do_one_spec
 
-printf_function *_custom_printf_handler[MAX_USER_SPEC];
+printf_function _custom_printf_handler[MAX_USER_SPEC];
 
 extern void _store_inttype(void *dest, int desttype, uintmax_t val);
 extern uintmax_t _load_inttype(int desttype, const void *src, int uflag);
@@ -1150,7 +1168,7 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 				if (ppfs->conv_num == CONV_X) {
 					prefix_num = PREFIX_UPR_X;
 				}
-				if ((ppfs->conv_num == CONV_o) && (numfill < slen)) {
+				if ((ppfs->conv_num == CONV_o) && (numfill <= slen)) {
 					numfill = ((*s == '0') ? 1 : slen + 1);
 				}
 			}
@@ -1207,7 +1225,7 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 			}
 #ifdef __STDIO_PRINTF_M_SUPPORT
 		} else if (ppfs->conv_num == CONV_m) {
-			s = _stdio_strerror_r(errno, buf, sizeof(buf));
+			s = _glibc_strerror_r(errno, buf, sizeof(buf));
 			goto SET_STRING_LEN;
 #endif
 		} else {
@@ -1263,6 +1281,7 @@ int _do_one_spec(FILE * __restrict stream, register ppfs_t *ppfs, int *count)
 #endif
 /**********************************************************************/
 #ifdef L_vsnprintf
+
 #ifdef __STDIO_BUFFERS
 int vsnprintf(char *__restrict buf, size_t size,
 			  const char * __restrict format, va_list arg)
@@ -1273,7 +1292,7 @@ int vsnprintf(char *__restrict buf, size_t size,
 #ifdef __STDIO_GETC_MACRO
 	f.bufgetc =
 #endif
-	f.bufrpos = f.bufwpos = f.bufstart = buf;
+	f.bufpos = f.bufread = f.bufstart = buf;
 
 	if (size > SIZE_MAX - (size_t) buf) {
 		size = SIZE_MAX - (size_t) buf;
@@ -1294,16 +1313,21 @@ int vsnprintf(char *__restrict buf, size_t size,
 	f.filedes = -2;				/* for debugging */
 	f.modeflags = (__FLAG_NARROW|__FLAG_WRITEONLY|__FLAG_WRITING);
 
+#ifdef __STDIO_MBSTATE
+	__INIT_MBSTATE(&(f.state));
+#endif /* __STDIO_MBSTATE */
+
 #ifdef __STDIO_THREADSAFE
+	f.user_locking = 0;
 	__stdio_init_mutex(&f.lock);
 #endif
 
 	rv = vfprintf(&f, format, arg);
 	if (size) {
-		if (f.bufwpos == f.bufend) {
-			--f.bufwpos;
+		if (f.bufpos == f.bufend) {
+			--f.bufpos;
 		}
-		*f.bufwpos = 0;
+		*f.bufpos = 0;
 	}
 	return rv;
 }
@@ -1319,10 +1343,11 @@ typedef struct {
 
 #define COOKIE ((__snpf_cookie *) cookie)
 
-static ssize_t snpf_write(void *cookie, const char *buf, size_t bufsize)
+static ssize_t snpf_write(register void *cookie, const char *buf,
+						  size_t bufsize)
 {
 	size_t count;
-	char *p;
+	register char *p;
 
 	/* Note: bufsize < SSIZE_MAX because of _stdio_WRITE. */
 
@@ -1368,7 +1393,12 @@ int vsnprintf(char *__restrict buf, size_t size,
 	f.filedes = -1;				/* For debugging. */
 	f.modeflags = (__FLAG_NARROW|__FLAG_WRITEONLY|__FLAG_WRITING);
 
+#ifdef __STDIO_MBSTATE
+	__INIT_MBSTATE(&(f.state));
+#endif /* __STDIO_MBSTATE */
+
 #ifdef __STDIO_THREADSAFE
+	f.user_locking = 0;
 	__stdio_init_mutex(&f.lock);
 #endif
 
@@ -1384,6 +1414,7 @@ int vsnprintf(char *__restrict buf, size_t size,
 #endif
 /**********************************************************************/
 #ifdef L_vdprintf
+
 int vdprintf(int filedes, const char * __restrict format, va_list arg)
 {
 	FILE f;
@@ -1394,7 +1425,7 @@ int vdprintf(int filedes, const char * __restrict format, va_list arg)
 #ifdef __STDIO_GETC_MACRO
 	f.bufgetc =
 #endif
-	f.bufrpos = f.bufwpos = f.bufstart = buf;
+	f.bufpos = f.bufread = f.bufstart = buf;
 #ifdef __STDIO_PUTC_MACRO
 	f.bufputc = 
 #endif
@@ -1410,7 +1441,12 @@ int vdprintf(int filedes, const char * __restrict format, va_list arg)
 	f.filedes = filedes;
 	f.modeflags = (__FLAG_NARROW|__FLAG_WRITEONLY|__FLAG_WRITING);
 
+#ifdef __STDIO_MBSTATE
+	__INIT_MBSTATE(&(f.state));
+#endif /* __STDIO_MBSTATE */
+
 #ifdef __STDIO_THREADSAFE
+	f.user_locking = 0;
 	__stdio_init_mutex(&f.lock);
 #endif
 
@@ -1418,6 +1454,7 @@ int vdprintf(int filedes, const char * __restrict format, va_list arg)
 
 	return fflush(&f) ? -1 : rv;
 }
+
 #endif
 /**********************************************************************/
 #ifdef L_vasprintf
@@ -1486,6 +1523,7 @@ int vsprintf(char *__restrict buf, const char * __restrict format,
 #endif
 /**********************************************************************/
 #ifdef L_fprintf
+
 int fprintf(FILE * __restrict stream, const char * __restrict format, ...)
 {
 	va_list arg;
@@ -1497,6 +1535,7 @@ int fprintf(FILE * __restrict stream, const char * __restrict format, ...)
 
 	return rv;
 }
+
 #endif
 /**********************************************************************/
 #ifdef L_snprintf
@@ -1521,6 +1560,7 @@ int snprintf(char *__restrict buf, size_t size,
 #endif
 /**********************************************************************/
 #ifdef L_dprintf
+
 int dprintf(int filedes, const char * __restrict format, ...)
 {
 	va_list arg;
@@ -1532,6 +1572,7 @@ int dprintf(int filedes, const char * __restrict format, ...)
 
 	return rv;
 }
+
 #endif
 /**********************************************************************/
 #ifdef L_asprintf
@@ -1690,14 +1731,14 @@ int sprintf(char *__restrict buf, const char * __restrict format, ...)
  */
 #define DIGITS_PER_BLOCK     9
 
-#if UINT_MAX >= 4294967295UL
+#if INT_MAX >= 2147483647L
 #define DIGIT_BLOCK_TYPE     int
 #define DB_FMT               "%.*d"
-#elif LONG_MAX >= 4294967295UL
+#elif LONG_MAX >= 2147483647L
 #define DIGIT_BLOCK_TYPE     long
 #define DB_FMT               "%.*ld"
 #else
-#error need at least 32 bit longs
+#warning need at least 32 bit longs
 #endif
 
 /* Maximum number of calls to fnprintf to output double. */
@@ -1859,6 +1900,11 @@ size_t _dtostr(FILE * fp, long double x, struct printf_info *info)
 
 	if (mode == 'f') {
 		round += exp;
+		if (round < -1) {
+			memset(buf, '0', MAX_DIGITS);
+		    exp = -1;
+		    round = -1;
+		}
 	}
 
 	s = buf;
@@ -2059,7 +2105,8 @@ void _store_inttype(register void *dest, int desttype, uintmax_t val)
 /**********************************************************************/
 #ifdef L__load_inttype
 
-extern uintmax_t _load_inttype(int desttype, const void *src, int uflag)
+extern uintmax_t _load_inttype(int desttype, register const void *src,
+							   int uflag)
 {
 	if (uflag >= 0) {			/* unsigned */
 #if LONG_MAX != INT_MAX
