@@ -62,6 +62,13 @@ typedef unsigned char	__string_uchar_t;
  * have to be changed! */
 
 #define _SYS_NERR			125
+#if defined(__mips__) || defined(__sparc__)
+/* sparce and mips have an extra error entry, as EDEADLK and EDEADLOCK have
+ * different meanings on those platforms. */
+#undef _SYS_NERR
+#define _SYS_NERR			126
+#endif
+
 #define _SYS_ERRMSG_MAXLEN	 50
 
 extern const char _string_syserrmsgs[];
@@ -213,6 +220,12 @@ const char _string_syserrmsgs[] = {
 	/* 122: 2852, 20 */ "Disk quota exceeded\0"
 	/* 123: 2872, 16 */ "No medium found\0"
 	/* 124: 2888, 18 */ "Wrong medium type"
+#if defined(__mips__) || defined(__sparc__)
+	"\0"
+	/* 125: 2906, 28 */ "File locking deadlock error"
+#endif
+	/* Note: for mips we are ignoring ECANCELED since glibc doesn't have a
+	 * corresponsding message.*/
 };
 
 #endif
@@ -221,138 +234,160 @@ const char _string_syserrmsgs[] = {
 
 link_warning(_sys_errlist, "sys_nerr and sys_errlist are obsolete and uClibc support for them (in at least some configurations) will probably be unavailable in the near future.")
 
-int sys_nerr = _SYS_NERR;
-
 const char *const sys_errlist[] = {
-	_string_syserrmsgs + 0,
-	_string_syserrmsgs + 8,
-	_string_syserrmsgs + 32,
-	_string_syserrmsgs + 58,
-	_string_syserrmsgs + 74,
-	_string_syserrmsgs + 98,
-	_string_syserrmsgs + 117,
-	_string_syserrmsgs + 143,
-	_string_syserrmsgs + 166,
-	_string_syserrmsgs + 184,
-	_string_syserrmsgs + 204,
-	_string_syserrmsgs + 223,
-	_string_syserrmsgs + 256,
-	_string_syserrmsgs + 279,
-	_string_syserrmsgs + 297,
-	_string_syserrmsgs + 309,
-	_string_syserrmsgs + 331,
-	_string_syserrmsgs + 355,
-	_string_syserrmsgs + 367,
-	_string_syserrmsgs + 393,
-	_string_syserrmsgs + 408,
-	_string_syserrmsgs + 424,
-	_string_syserrmsgs + 439,
-	_string_syserrmsgs + 456,
-	_string_syserrmsgs + 486,
-	_string_syserrmsgs + 506,
-	_string_syserrmsgs + 537,
-	_string_syserrmsgs + 552,
-	_string_syserrmsgs + 567,
-	_string_syserrmsgs + 591,
-	_string_syserrmsgs + 604,
-	_string_syserrmsgs + 626,
-	_string_syserrmsgs + 641,
-	_string_syserrmsgs + 653,
-	_string_syserrmsgs + 686,
-	_string_syserrmsgs + 716,
-	_string_syserrmsgs + 742,
-	_string_syserrmsgs + 761,
-	_string_syserrmsgs + 780,
-	_string_syserrmsgs + 805,
-	_string_syserrmsgs + 825,
+	[0] =				_string_syserrmsgs + 0,
+	[EPERM] =			_string_syserrmsgs + 8,
+	[ENOENT] =			_string_syserrmsgs + 32,
+	[ESRCH] =			_string_syserrmsgs + 58,
+	[EINTR] =			_string_syserrmsgs + 74,
+	[EIO] =				_string_syserrmsgs + 98,
+	[ENXIO] =			_string_syserrmsgs + 117,
+	[E2BIG] =			_string_syserrmsgs + 143,
+	[ENOEXEC] =			_string_syserrmsgs + 166,
+	[EBADF] =			_string_syserrmsgs + 184,
+	[ECHILD] =			_string_syserrmsgs + 204,
+	[EAGAIN] =			_string_syserrmsgs + 223,
+	[ENOMEM] =			_string_syserrmsgs + 256,
+	[EACCES] =			_string_syserrmsgs + 279,
+	[EFAULT] =			_string_syserrmsgs + 297,
+	[ENOTBLK] =			_string_syserrmsgs + 309,
+	[EBUSY] =			_string_syserrmsgs + 331,
+	[EEXIST] =			_string_syserrmsgs + 355,
+	[EXDEV] =			_string_syserrmsgs + 367,
+	[ENODEV] =			_string_syserrmsgs + 393,
+	[ENOTDIR] =			_string_syserrmsgs + 408,
+	[EISDIR] =			_string_syserrmsgs + 424,
+	[EINVAL] =			_string_syserrmsgs + 439,
+	[ENFILE] =			_string_syserrmsgs + 456,
+	[EMFILE] =			_string_syserrmsgs + 486,
+	[ENOTTY] =			_string_syserrmsgs + 506,
+	[ETXTBSY] =			_string_syserrmsgs + 537,
+	[EFBIG] =			_string_syserrmsgs + 552,
+	[ENOSPC] =			_string_syserrmsgs + 567,
+	[ESPIPE] =			_string_syserrmsgs + 591,
+	[EROFS] =			_string_syserrmsgs + 604,
+	[EMLINK] =			_string_syserrmsgs + 626,
+	[EPIPE] =			_string_syserrmsgs + 641,
+	[EDOM] =			_string_syserrmsgs + 653,
+	[ERANGE] =			_string_syserrmsgs + 686,
+	[EDEADLK] =			_string_syserrmsgs + 716,
+	[ENAMETOOLONG] =	_string_syserrmsgs + 742,
+	[ENOLCK] =			_string_syserrmsgs + 761,
+	[ENOSYS] =			_string_syserrmsgs + 780,
+	[ENOTEMPTY] =		_string_syserrmsgs + 805,
+	[ELOOP] =			_string_syserrmsgs + 825,
 	/*  	_string_syserrmsgs + 859, */
-	NULL,						/* glibc compatiblity :-( */
-	_string_syserrmsgs + 860,
-	_string_syserrmsgs + 887,
-	_string_syserrmsgs + 906,
-	_string_syserrmsgs + 934,
-	_string_syserrmsgs + 959,
-	_string_syserrmsgs + 974,
-	_string_syserrmsgs + 988,
-	_string_syserrmsgs + 1013,
-	_string_syserrmsgs + 1042,
-	_string_syserrmsgs + 1069,
-	_string_syserrmsgs + 1084,
-	_string_syserrmsgs + 1101,
-	_string_syserrmsgs + 1128,
-	_string_syserrmsgs + 1142,
-	_string_syserrmsgs + 1151,
-	_string_syserrmsgs + 1172,
+	[ENOMSG] =			_string_syserrmsgs + 860,
+	[EIDRM] =			_string_syserrmsgs + 887,
+	[ECHRNG] =			_string_syserrmsgs + 906,
+	[EL2NSYNC] =		_string_syserrmsgs + 934,
+	[EL3HLT] =			_string_syserrmsgs + 959,
+	[EL3RST] =			_string_syserrmsgs + 974,
+	[ELNRNG] =			_string_syserrmsgs + 988,
+	[EUNATCH] =			_string_syserrmsgs + 1013,
+	[ENOCSI] =			_string_syserrmsgs + 1042,
+	[EL2HLT] =			_string_syserrmsgs + 1069,
+	[EBADE] =			_string_syserrmsgs + 1084,
+	[EBADR] =			_string_syserrmsgs + 1101,
+	[EXFULL] =			_string_syserrmsgs + 1128,
+	[ENOANO] =			_string_syserrmsgs + 1142,
+	[EBADRQC] =			_string_syserrmsgs + 1151,
+	[EBADSLT] =			_string_syserrmsgs + 1172,
 	/*  	_string_syserrmsgs + 1185, */
-	NULL,						/* glibc compatiblity :-( */
-	_string_syserrmsgs + 1186,
-	_string_syserrmsgs + 1207,
-	_string_syserrmsgs + 1227,
-	_string_syserrmsgs + 1245,
-	_string_syserrmsgs + 1259,
-	_string_syserrmsgs + 1284,
-	_string_syserrmsgs + 1314,
-	_string_syserrmsgs + 1336,
-	_string_syserrmsgs + 1353,
-	_string_syserrmsgs + 1375,
-	_string_syserrmsgs + 1391,
-	_string_syserrmsgs + 1405,
-	_string_syserrmsgs + 1433,
-	_string_syserrmsgs + 1448,
-	_string_syserrmsgs + 1467,
-	_string_syserrmsgs + 1486,
-	_string_syserrmsgs + 1498,
-	_string_syserrmsgs + 1536,
-	_string_syserrmsgs + 1563,
-	_string_syserrmsgs + 1592,
-	_string_syserrmsgs + 1615,
-	_string_syserrmsgs + 1654,
-	_string_syserrmsgs + 1691,
-	_string_syserrmsgs + 1723,
-	_string_syserrmsgs + 1771,
-	_string_syserrmsgs + 1809,
-	_string_syserrmsgs + 1859,
-	_string_syserrmsgs + 1903,
-	_string_syserrmsgs + 1922,
-	_string_syserrmsgs + 1937,
-	_string_syserrmsgs + 1968,
-	_string_syserrmsgs + 1997,
-	_string_syserrmsgs + 2014,
-	_string_syserrmsgs + 2045,
-	_string_syserrmsgs + 2068,
-	_string_syserrmsgs + 2091,
-	_string_syserrmsgs + 2117,
-	_string_syserrmsgs + 2141,
-	_string_syserrmsgs + 2171,
-	_string_syserrmsgs + 2212,
-	_string_syserrmsgs + 2235,
-	_string_syserrmsgs + 2267,
-	_string_syserrmsgs + 2283,
-	_string_syserrmsgs + 2306,
-	_string_syserrmsgs + 2342,
-	_string_syserrmsgs + 2375,
-	_string_syserrmsgs + 2400,
-	_string_syserrmsgs + 2426,
-	_string_syserrmsgs + 2466,
-	_string_syserrmsgs + 2502,
-	_string_syserrmsgs + 2548,
-	_string_syserrmsgs + 2583,
-	_string_syserrmsgs + 2604,
-	_string_syserrmsgs + 2623,
-	_string_syserrmsgs + 2636,
-	_string_syserrmsgs + 2653,
-	_string_syserrmsgs + 2683,
-	_string_syserrmsgs + 2709,
-	_string_syserrmsgs + 2731,
-	_string_syserrmsgs + 2756,
-	_string_syserrmsgs + 2784,
-	_string_syserrmsgs + 2814,
-	_string_syserrmsgs + 2835,
-	_string_syserrmsgs + 2852,
-	_string_syserrmsgs + 2872,
-	_string_syserrmsgs + 2888,
+	[EBFONT] =			_string_syserrmsgs + 1186,
+	[ENOSTR] =			_string_syserrmsgs + 1207,
+	[ENODATA] =			_string_syserrmsgs + 1227,
+	[ETIME] =			_string_syserrmsgs + 1245,
+	[ENOSR] =			_string_syserrmsgs + 1259,
+	[ENONET] =			_string_syserrmsgs + 1284,
+	[ENOPKG] =			_string_syserrmsgs + 1314,
+	[EREMOTE] =			_string_syserrmsgs + 1336,
+	[ENOLINK] =			_string_syserrmsgs + 1353,
+	[EADV] =			_string_syserrmsgs + 1375,
+	[ESRMNT] =			_string_syserrmsgs + 1391,
+	[ECOMM] =			_string_syserrmsgs + 1405,
+	[EPROTO] =			_string_syserrmsgs + 1433,
+	[EMULTIHOP] =		_string_syserrmsgs + 1448,
+	[EDOTDOT] =			_string_syserrmsgs + 1467,
+	[EBADMSG] =			_string_syserrmsgs + 1486,
+	[EOVERFLOW] =		_string_syserrmsgs + 1498,
+	[ENOTUNIQ] =		_string_syserrmsgs + 1536,
+	[EBADFD] =			_string_syserrmsgs + 1563,
+	[EREMCHG] =			_string_syserrmsgs + 1592,
+	[ELIBACC] =			_string_syserrmsgs + 1615,
+	[ELIBBAD] =			_string_syserrmsgs + 1654,
+	[ELIBSCN] =			_string_syserrmsgs + 1691,
+	[ELIBMAX] =			_string_syserrmsgs + 1723,
+	[ELIBEXEC] =		_string_syserrmsgs + 1771,
+	[EILSEQ] =			_string_syserrmsgs + 1809,
+	[ERESTART] =		_string_syserrmsgs + 1859,
+	[ESTRPIPE] =		_string_syserrmsgs + 1903,
+	[EUSERS] =			_string_syserrmsgs + 1922,
+	[ENOTSOCK] =		_string_syserrmsgs + 1937,
+	[EDESTADDRREQ] =	_string_syserrmsgs + 1968,
+	[EMSGSIZE] =		_string_syserrmsgs + 1997,
+	[EPROTOTYPE] =		_string_syserrmsgs + 2014,
+	[ENOPROTOOPT] =		_string_syserrmsgs + 2045,
+	[EPROTONOSUPPORT] =	_string_syserrmsgs + 2068,
+	[ESOCKTNOSUPPORT] =	_string_syserrmsgs + 2091,
+	[EOPNOTSUPP] =		_string_syserrmsgs + 2117,
+	[EPFNOSUPPORT] =	_string_syserrmsgs + 2141,
+	[EAFNOSUPPORT] =	_string_syserrmsgs + 2171,
+	[EADDRINUSE] =		_string_syserrmsgs + 2212,
+	[EADDRNOTAVAIL] =	_string_syserrmsgs + 2235,
+	[ENETDOWN] =		_string_syserrmsgs + 2267,
+	[ENETUNREACH] =		_string_syserrmsgs + 2283,
+	[ENETRESET] =		_string_syserrmsgs + 2306,
+	[ECONNABORTED] =	_string_syserrmsgs + 2342,
+	[ECONNRESET] =		_string_syserrmsgs + 2375,
+	[ENOBUFS] =			_string_syserrmsgs + 2400,
+	[EISCONN] =			_string_syserrmsgs + 2426,
+	[ENOTCONN] =		_string_syserrmsgs + 2466,
+	[ESHUTDOWN] =		_string_syserrmsgs + 2502,
+	[ETOOMANYREFS] =	_string_syserrmsgs + 2548,
+	[ETIMEDOUT] =		_string_syserrmsgs + 2583,
+	[ECONNREFUSED] =	_string_syserrmsgs + 2604,
+	[EHOSTDOWN] =		_string_syserrmsgs + 2623,
+	[EHOSTUNREACH] =	_string_syserrmsgs + 2636,
+	[EALREADY] =		_string_syserrmsgs + 2653,
+	[EINPROGRESS] =		_string_syserrmsgs + 2683,
+	[ESTALE] =			_string_syserrmsgs + 2709,
+	[EUCLEAN] =			_string_syserrmsgs + 2731,
+	[ENOTNAM] =			_string_syserrmsgs + 2756,
+	[ENAVAIL] =			_string_syserrmsgs + 2784,
+	[EISNAM] =			_string_syserrmsgs + 2814,
+	[EREMOTEIO] =		_string_syserrmsgs + 2835,
+	[EDQUOT] =			_string_syserrmsgs + 2852,
+	[ENOMEDIUM] =		_string_syserrmsgs + 2872,
+	[EMEDIUMTYPE] =		_string_syserrmsgs + 2888,
+
+#if EDEADLOCK != EDEADLK
+	[EDEADLOCK] =		_string_syserrmsgs + 2906,
+#endif
+
+#if EWOULDBLOCK != EAGAIN
+#warning EWOULDBLOCK does not equal EAGAIN
+#endif
+
+	/* For now, ignore the other arch-specific errors.  glibc only maps EPROCLIM. */
+
+	/* some other mips errors */
+#ifdef ECANCELED
+#endif
+#ifdef EINIT
+#endif
+#ifdef EREMDEV
+#endif
+
+	/* some other sparc errors */
+#ifdef EPROCLIM
+#endif
+#ifdef ERREMOTE
+#endif
 };
 
+int sys_nerr = sizeof(sys_errlist)/sizeof(sys_errlist[0]);
+
 #endif
 /**********************************************************************/
 #ifdef L_wmemcpy
@@ -1198,7 +1233,9 @@ char *strerror(int errnum)
 {
     static char buf[_SYS_ERRMSG_MAXLEN];
 
-    return (_susv3_strerror_r(errnum, buf, sizeof(buf)) == 0) ? buf : NULL;
+	_susv3_strerror_r(errnum, buf, sizeof(buf));
+
+	return buf;
 }
 
 #endif
@@ -1207,6 +1244,146 @@ char *strerror(int errnum)
 /**********************************************************************/
 #ifdef L__susv3_strerror_r
 
+#if defined(__alpha__) || defined(__mips__) || defined(__sparc__)
+
+static const unsigned char estridx[] = {
+	0,							/* success is always 0 */
+	EPERM,
+	ENOENT,
+	ESRCH,
+	EINTR,
+	EIO,
+	ENXIO,
+	E2BIG,
+	ENOEXEC,
+	EBADF,
+	ECHILD,
+	EAGAIN,
+	ENOMEM,
+	EACCES,
+	EFAULT,
+	ENOTBLK,
+	EBUSY,
+	EEXIST,
+	EXDEV,
+	ENODEV,
+	ENOTDIR,
+	EISDIR,
+	EINVAL,
+	ENFILE,
+	EMFILE,
+	ENOTTY,
+	ETXTBSY,
+	EFBIG,
+	ENOSPC,
+	ESPIPE,
+	EROFS,
+	EMLINK,
+	EPIPE,
+	EDOM,
+	ERANGE,
+	EDEADLK,
+	ENAMETOOLONG,
+	ENOLCK,
+	ENOSYS,
+	ENOTEMPTY,
+	ELOOP,
+	0,
+	ENOMSG,
+	EIDRM,
+	ECHRNG,
+	EL2NSYNC,
+	EL3HLT,
+	EL3RST,
+	ELNRNG,
+	EUNATCH,
+	ENOCSI,
+	EL2HLT,
+	EBADE,
+	EBADR,
+	EXFULL,
+	ENOANO,
+	EBADRQC,
+	EBADSLT,
+	0,
+	EBFONT,
+	ENOSTR,
+	ENODATA,
+	ETIME,
+	ENOSR,
+	ENONET,
+	ENOPKG,
+	EREMOTE,
+	ENOLINK,
+	EADV,
+	ESRMNT,
+	ECOMM,
+	EPROTO,
+	EMULTIHOP,
+	EDOTDOT,
+	EBADMSG,
+	EOVERFLOW,
+	ENOTUNIQ,
+	EBADFD,
+	EREMCHG,
+	ELIBACC,
+	ELIBBAD,
+	ELIBSCN,
+	ELIBMAX,
+	ELIBEXEC,
+	EILSEQ,
+	ERESTART,
+	ESTRPIPE,
+	EUSERS,
+	ENOTSOCK,
+	EDESTADDRREQ,
+	EMSGSIZE,
+	EPROTOTYPE,
+	ENOPROTOOPT,
+	EPROTONOSUPPORT,
+	ESOCKTNOSUPPORT,
+	EOPNOTSUPP,
+	EPFNOSUPPORT,
+	EAFNOSUPPORT,
+	EADDRINUSE,
+	EADDRNOTAVAIL,
+	ENETDOWN,
+	ENETUNREACH,
+	ENETRESET,
+	ECONNABORTED,
+	ECONNRESET,
+	ENOBUFS,
+	EISCONN,
+	ENOTCONN,
+	ESHUTDOWN,
+	ETOOMANYREFS,
+	ETIMEDOUT,
+	ECONNREFUSED,
+	EHOSTDOWN,
+	EHOSTUNREACH,
+	EALREADY,
+	EINPROGRESS,
+	ESTALE,
+	EUCLEAN,
+	ENOTNAM,
+	ENAVAIL,
+	EISNAM,
+	EREMOTEIO,
+#ifdef __mips__
+	0,							/* mips has an outrageous value for this... */
+#else
+	EDQUOT,
+#endif
+	ENOMEDIUM,
+	EMEDIUMTYPE,
+#if defined(__mips__) || defined(__sparc__)
+	EDEADLOCK,
+#endif
+};
+
+#endif
+
+
 int _susv3_strerror_r(int errnum, char *strerrbuf, size_t buflen)
 {
     register char *s;
@@ -1218,12 +1395,32 @@ int _susv3_strerror_r(int errnum, char *strerrbuf, size_t buflen)
 
     retval = EINVAL;
 
-    if (((unsigned int) errnum) < _SYS_NERR) {
+
+#if defined(__alpha__) || defined(__mips__) || defined(__sparc__)
+	/* Need to translate errno to string index. */
+	for (i = 0 ; i < sizeof(estridx)/sizeof(estridx[0]) ; i++) {
+		if (estridx[i] == errnum) {
+			goto GOT_ESTRIDX;
+		}
+	}
+	i = INT_MAX;	/* Failed, but may need to check mips special case. */
+#ifdef __mips__
+	if (errnum == EDQUOT) {	/* Deal with large EDQUOT value on mips */
+		i = 122;
+	}
+#endif /* __mips__ */
+ GOT_ESTRIDX:
+#else
+	/* No errno to string index translation needed. */
+	i = errnum;
+#endif
+
+    if (((unsigned int) i) < _SYS_NERR) {
 		/* Trade time for space.  This function should rarely be called
 		 * so rather than keeping an array of pointers for the different
 		 * messages, just run through the buffer until we find the
 		 * correct string. */
-		for (s = (char *) _string_syserrmsgs, i = errnum ; i ; ++s) {
+		for (s = (char *) _string_syserrmsgs ; i ; ++s) {
 			if (!*s) {
 				--i;
 			}
