@@ -26,10 +26,11 @@
 #endif
 
 #ifdef __UCLIBC_HAVE_LFS__
+extern int __libc_open (__const char *file, int oflag, mode_t mode);
 
 /* Open FILE with access OFLAG.  If OFLAG includes O_CREAT,
    a third argument is the file protection.  */
-int open64 (const char *file, int oflag, ...)
+int __libc_open64 (const char *file, int oflag, ...)
 {
   int mode = 0;
 
@@ -41,8 +42,7 @@ int open64 (const char *file, int oflag, ...)
       va_end (arg);
     }
 
-  return open(file, oflag | O_LARGEFILE, mode);
+  return __libc_open(file, oflag | O_LARGEFILE, mode);
 }
-
+weak_alias (__libc_open64, open64);
 #endif /* __UCLIBC_HAVE_LFS__ */
-
